/**
 * @file
 * RefreshLess progress bar.
 *
 * @see https://www.drupal.org/project/refreshless/issues/3488464
 *   Issue for adapting this out of Turbo.
 *
 * @see https://github.com/hotwired/turbo/blob/ea54ae5ad4b6b28cb62ccd62951352641ed08293/src/core/drive/progress_bar.js
 *   Originally adapted from the Turbo progress bar.
 */
((Drupal) => {

  'use strict';

  Drupal.theme.refreshlessProgressBar = (
    elementType   = 'div',
    elementClass  = 'refreshless-document-progress-bar',
  ) => {

    const element = document.createElement(elementType);

    element.classList.add(elementClass);

    // This provides the ProgressBar JavaScript class the configured class name
    // to build modifier classes from.
    element.refreshlessProgressBarClass = elementClass;

    return element;

  }

  /**
   * CSS custom property base name.
   *
   * @type {String}
   */
  const customPropertyBase = '--refreshless-progress-bar';

  /**
   * Name of the CSS custom property containing the transition duration.
   *
   * The value will be a time in ms.
   *
   * @type {String}
   */
  const durationCustomProperty = `${customPropertyBase}-transition-duration`;

  /**
   * Name of the CSS custom property containing the current progress bar value.
   *
   * The value will be float between 0 and 1, inclusive.
   *
   * @type {String}
   */
  const valueCustomProperty = `${customPropertyBase}-value`;

  /**
   * The <html> element.
   *
   * @type {HTMLHtmlElement}
   */
  const html = document.documentElement;

  /**
   * Represents a progress bar.
   */
  class ProgressBar {

    /**
     * The progress bar HTML element.
     *
     * @type {HTMLElement}
     */
    #element;

    /**
     * Whether the progress bar is currently in the processing of hiding.
     *
     * @type {Boolean}
     */
    #hiding = false;

    /**
     * The current value of the progress bar, from 0 to 1, inclusive.
     *
     * @type {Number}
     */
    #value = 0;

    /**
     * Whether the progress bar is currently visible.
     *
     * @type {Boolean}
     */
    #visible = false;

    /**
     * The trickle interval ID, or null if one is not active.
     *
     * @type {Number|null}
     */
    #trickleInterval = null;

    /**
     * The progress bar transition duration, in milliseconds.
     *
     * @type {Number}
     */
    #transitionDuration = 300;

    constructor() {

      this.#element = Drupal.theme('refreshlessProgressBar');

      this.setValue(0, false);

    }

    /**
     * Show the progress bar if not already visible.
     */
    show() {

      if (this.#visible === true) {
        return;
      }

      // @todo Make cancellable.
      html.dispatchEvent(new CustomEvent(
        'refreshless:progress-bar-before-show',
      ));

      this.#visible = true;
      this.install();
      this.startTrickling();

      html.dispatchEvent(new CustomEvent('refreshless:progress-bar-shown'));

    }

    /**
     * Hide the progress bar if visible and not already hiding.
     *
     * @return {Promise}
     */
    hide() {

      if (!(this.#visible === true && this.#hiding === false)) {
        return Promise.resolve();
      }

      // @todo Make cancellable.
      html.dispatchEvent(new CustomEvent(
        'refreshless:progress-bar-before-hide',
      ));

      this.#hiding = true;

      this.stopTrickling();

      return new Promise(async (resolve, reject) => {

        this.#setInactive();

        await new Promise((resolve, reject) => {

          window.setTimeout(resolve, this.#transitionDuration * 1.5);

        });

        this.uninstall();
        this.#visible = false;
        this.#hiding = false;

        html.dispatchEvent(new CustomEvent('refreshless:progress-bar-hidden'));

        resolve();

      });

    }

    /**
     * Set the progress bar as active, causing CSS to transition it in.
     *
     * @param {Boolean} dispatch
     */
    #setActive(dispatch = true) {

      this.#element.classList.add(
        `${this.#element.refreshlessProgressBarClass}--active`,
      );

      if (dispatch === false) {
        return;
      }

      html.dispatchEvent(new CustomEvent('refreshless:progress-bar-active'));

    }

    /**
     * Set the progress bar as inactive, causing CSS to transition it out.
     *
     * @param {Boolean} dispatch
     *   Whether to dispatch an event. Defaults to true.
     */
    #setInactive(dispatch = true) {

      this.#element.classList.remove(
        `${this.#element.refreshlessProgressBarClass}--active`,
      );

      if (dispatch === false) {
        return;
      }

      html.dispatchEvent(new CustomEvent('refreshless:progress-bar-inactive'));

    }

    /**
     * Explicitly set the progress bar to a value.
     *
     * @param {Number} value
     *   A number between 0 and 1, inclusive.
     *
     * @param {Boolean} dispatch
     *   Whether to dispatch an event. Defaults to true.
     *
     * @throws If value is NaN, or if the value is less than 0 or greater than
     *   1.
     */
    setValue(value, dispatch = true) {

      if (Number.isNaN(value)) {
        throw `Progress bar value must be a number! Got: "${typeof value}"`;
      }

      if (value < 0 || value > 1) {
        throw `Progress bar value must be an integer or float between 0 and 1! Got: "${value}"`;
      }

      const valueChanged = (this.#value !== value);

      const oldValue = this.#value;

      this.#value = value;

      return this.refresh().then(() => {

        if (dispatch === false || valueChanged === false) {
          return;
        }

        html.dispatchEvent(new CustomEvent(
          'refreshless:progress-bar-value-changed', {detail: {
            value:    this.#value,
            oldValue: oldValue,
          }}
        ));

      });

    }

    /**
     * Install the progress bar in the document and set various properties.
     */
    install() {

      html.style.setProperty(
        durationCustomProperty, `${this.#transitionDuration}ms`,
      );

      html.style.setProperty(valueCustomProperty, 0);

      this.#setInactive(false);

      document.body.insertAdjacentElement('beforebegin', this.#element);

      this.refresh().then(async () => {

        await new Promise(requestAnimationFrame);

        this.#setActive();

      });

    }

    /**
     * Uninstall the progress bar from the document and remove properties.
     */
    uninstall() {

      if (this.#element.parentNode) {
        this.#element.remove();
      }

      html.style.removeProperty(durationCustomProperty);
      html.style.removeProperty(valueCustomProperty);

    }

    /**
     * Start the trickling animation.
     */
    startTrickling() {

      if (this.#trickleInterval !== null) {
        return;
      }

      this.#trickleInterval = window.setInterval(
        this.trickle, this.#transitionDuration,
      );

    }

    /**
     * Stop the trickling animation.
     */
    stopTrickling() {

      window.clearInterval(this.#trickleInterval);

      this.#trickleInterval = null;

    }

    /**
     * Trickle animation interval callback.
     *
     * This generates a random value to give the trickle the irregular movement.
     */
    trickle = () => {
      this.setValue(Math.min(1, this.#value + Math.random() / 100));
    }

    /**
     * Update the progress bar element's value with the current value.
     */
    async refresh() {

      // Don't write the custom property if not visible or if currently hiding.
      if (!(this.#visible === true && this.#hiding === false)) {
        return;
      }

      await new Promise(requestAnimationFrame);

      html.style.setProperty(valueCustomProperty, this.#value);

    }

    /**
     * Finish/complete the progress bar by setting to 100% and start hiding it.
     *
     * @return {Promise}
     */
    finish() {

      html.dispatchEvent(new CustomEvent('refreshless:progress-bar-finish'));

      return this.setValue(1).then(() => this.hide());

    }

    /**
     * Get the current value of the progress bar.
     *
     * @return {Number}
     */
    get value() {
      return this.#value;
    }

    /**
     * Get the progress bar HTML element.
     *
     * @return {HTMLElement}
     */
    get element() {
      return this.#element;
    }

    /**
     * Get the progress bar transition value.
     *
     * @return {Number}
     */
    get transitionDuration() {
      return this.#transitionDuration;
    }

  }

  if (!('RefreshLess' in Drupal)) {
    Drupal.RefreshLess = {};
  }

  if (!('classes' in Drupal.RefreshLess)) {
    Drupal.RefreshLess.classes = {};
  }

  Drupal.RefreshLess.classes.ProgressBar = ProgressBar;

})(Drupal);
