(function (Drupal, drupalSettings) {

  'use strict';

  if (!('RefreshLess' in Drupal)) {
    Drupal.RefreshLess = {};
  }

  if (!('classes' in Drupal.RefreshLess)) {
    Drupal.RefreshLess.classes = {};
  }

  /**
   * RefreshLess Turbo behaviours attach and detach class.
   */
  Drupal.RefreshLess.classes.TurboBehaviours = class {

    /**
     * Array of <script> element Promises, each settled when they load or fail.
     *
     * Note that we probably don't need to remove any Promises from this,
     * because Promises that are already settled when they're passed to
     * Promise.allSettled() should just instantly count as settled to that
     * method.
     *
     * @type {Promise[]}
     */
    #scriptPromises = [];

    /**
     * <script> element MutationObserver.
     *
     * @type {MutationObserver}
     */
    #scriptObserver;

    constructor() {

      /**
       * Reference to the current instance.
       *
       * @type {Behaviours}
       */
      const that = this;

      /**
       * Reference to our #scriptObserverCallback private method.
       *
       * @type {Function}
       */
      const scriptObserverCallback = this.#scriptObserverCallback;

      this.#scriptObserver = new MutationObserver(function(mutations) {
        scriptObserverCallback.call(that, mutations);
      });

      /**
       * Reference to our <script> element MutationObserver.
       *
       * @type {MutationObserver}
       */
      const scriptObserver = this.#scriptObserver;

      document.documentElement.addEventListener('turbo:visit', function(event) {

        // Start observing when a Turbo visit starts. This event is triggered
        // before any changes are made to the DOM, including the <head>, so no
        // new <script> elements will have been merged in yet.
        scriptObserver.observe(document.documentElement, {
          childList:  true,
          subtree:    true,
        });

      });

    };

    /**
     * <script> element MutationObserver callback.
     *
     * @param {MutationRecord} mutations
     */
    #scriptObserverCallback(mutations) {

      for (let i = 0; i < mutations.length; i++) {

        if (!('addedNodes' in mutations[i])) {
          continue;
        }

        /**
         * Nodes added as part of this mutation record.
         *
         * @type {NodeList}
         */
        const addedNodes = mutations[i].addedNodes;

        for (let j = 0; j < addedNodes.length; j++) {

          // Ignore any non-<script> element nodes and the drupalSettings JSON
          // element.
          if (
            !(addedNodes[j] instanceof HTMLScriptElement) ||
            addedNodes[j].getAttribute(
              'data-drupal-selector',
            ) === 'drupal-settings-json'
          ) {
            continue;
          }

          this.#createScriptPromise(addedNodes[j]);

        }

      }

    };

    /**
     * Create a Promise that resolves or rejects when a <script> loads/errors.
     *
     * This creates a Promise for the <script> element with the resolve() and
     * reject() callbacks assigned as the load and error event handlers,
     * respectively, and pushes that Promise onto this.#scriptPromises.
     *
     * @param {HTMLScriptElement} element
     *   A <script> element add a load/error Promise for.
     */
    #createScriptPromise(element) {

      this.#scriptPromises.push(new Promise(function(resolve, reject) {
        element.addEventListener('load',  resolve,  {once: true});
        element.addEventListener('error', reject,   {once: true});
      }));

    }

    /**
     * Attach behaviours to the provided context element.
     *
     * @param {HTMLElement} context
     *   The context element to attach behaviours to.
     *
     * @return {Promise}
     *   A Promise that fulfills when Drupal.attachBehaviors() has been called.
     */
    attach(context) {

      /**
       * Reference to our <script> element MutationObserver.
       *
       * @type {MutationObserver}
       */
      const scriptObserver = this.#scriptObserver;

      // Process any pending mutations not yet processed by the callback.
      this.#scriptObserverCallback(scriptObserver.takeRecords());

      // Only when all of the Promises have settled - that is they've either
      // resolved or rejected, i.e. loaded or failed to load - do we attach
      // behaviours. We don't know ahead of time if one or more behaviours
      // being added by one of the newly added scripts depends on something
      // from one of the other newly added scripts, and because the scripts are
      // not being evalutated at first load in a predictable sequence but
      // dynamically later, this reduces the chances of undefined or reference
      // errors.
      return Promise.allSettled(this.#scriptPromises).then(function() {

        // Stop observing for new <script> elements at this point because our
        // job is done and we don't want to do unnecessary work and potentially
        // affect performance.
        scriptObserver.disconnect();

        console.debug('%cRefreshLess%c: attaching behaviours',
          'font-style: italic', 'font-style: normal',
        );

        Drupal.attachBehaviors(context, drupalSettings);

      });

    };

    /**
     * Detach behaviours from the provided context element.
     *
     * @param {HTMLElement} context
     *   The context element to detach behaviours from.
     *
     * @return {Promise}
     *   A Promise that fulfills when Drupal.detachBehaviors() has been called.
     */
    detach(context) {

      console.debug('%cRefreshLess%c: detaching behaviours',
        'font-style: italic', 'font-style: normal',
      );

      Drupal.detachBehaviors(context, drupalSettings, 'unload');

      return Promise.resolve();

    };

  }

})(Drupal, drupalSettings);
