(function (Drupal, drupalSettings, $) {

  'use strict';

  if (!('RefreshLess' in Drupal)) {
    Drupal.RefreshLess = {};
  }

  if (!('classes' in Drupal.RefreshLess)) {
    Drupal.RefreshLess.classes = {};
  }

  /**
   * RefreshLess Turbo drupalSettings updater class.
   */
  Drupal.RefreshLess.classes.TurboDrupalSettings = class {

    /**
     * The drupalSettings <script> element selector.
     *
     * Note that this doesn't include the leading '>' combinator for XSS
     * hardening.
     *
     * @type {String}
     */
    #selector = 'script[type="application/json"][data-drupal-selector="drupal-settings-json"]';

    constructor() {

      this.#bindEventHandlers();

    };

    /**
     * Bind our event handlers.
     */
    #bindEventHandlers() {

      /**
       * Reference to the current instance.
       *
       * @type {DrupalSettings}
       */
      const that = this;

      document.documentElement.addEventListener('turbo:render', function(
        event,
      ) {
        that.#renderHandler(event);
      });

    }

    /**
     * 'turbo:render' event handler.
     *
     * This removes any drupalSettings <script> elements Turbo will have merged
     * into the <head>.
     *
     * @param {CustomEvent} event
     *   The event object generated by Turbo.
     *
     * @see https://turbo.hotwired.dev/reference/events
     */
    #renderHandler(event) {

      // Don't do anything if this is a cached preview being shown.
      if (typeof $(event.target).attr('data-turbo-preview') !== 'undefined') {
        return;
      }

      $(event.target).find(`> head > ${this.#selector}`).remove();

    }

    /**
     * Update the drupalSettings object from JSON found in the DOM.
     *
     * @todo Should we instead discard the current settings object rather than
     *   merging?
     *
     * @throws Error
     *   If no settings element was found or if the settings element text could
     *   not be parsed into valid JSON.
     *
     * @return {Promise}
     *   A Promise that fulfills when drupalSettings has been updated.
     */
    update(documentElement) {

      /**
       * One or zero drupalSettings <script> elements in a jQuery collection.
       *
       * @type {jQuery}
       *
       * Note the leading '>' combinator; important for XSS hardening to only
       * match direct children rather than potential <script> elements
       * injected into content.
       *
       * @see core/misc/drupalSettingsLoader.js
       *   Explains the need for XSS hardening in the selector.
       */
      const $script = $(documentElement).find(
        `> body > ${this.#selector}`,
      ).first();

      if ($script.length === 0) {

        throw new Error(Drupal.t(
          'Could not find a Drupal settings <script> element!',
        ));

      }

      /**
       * Clone of the previous drupalSettings before merging.
       *
       * @type {Object}
       */
      const previousSettings = $.extend(true, {}, drupalSettings);

      /**
       * New drupalSettings values or null if they can't be read.
       *
       * @type {Object|null}
       */
      const newSettings = JSON.parse($script.text());

      if (newSettings === null) {

        throw new Error(Drupal.t(
          `Could not parse the drupalSettings JSON from the <script> element. Got: @text`,
            {'@text': $script.text()},
        ));

      }

      $.extend(true, drupalSettings, newSettings);

      return Promise.resolve({
        new:      newSettings,
        previous: previousSettings,
        merged:   drupalSettings,
      });

    };

  }

})(Drupal, drupalSettings, jQuery);
