(function (Drupal, drupalSettings, $, Cookies) {

  'use strict';

  /**
   * Name of the cookie we set to pass already loaded libraries to the back-end.
   *
   * @type {String}
   */
  const pageStateCookieName = 'refreshless-turbo-page-state';

  /**
   * Name of the cookie we set to indicate a reload to the back-end.
   *
   * @type {String}
   */
  const reloadReasonCookieName = 'refreshless-turbo-reload-reason';

  /**
   * Name of the header we add to indicate a RefreshLess Turbo request.
   *
   * @type {String}
   */
  const turboRequestHeaderName = 'x-refreshless-turbo';

  if (!('RefreshLess' in Drupal)) {
    Drupal.RefreshLess = {};
  }

  if (!('classes' in Drupal.RefreshLess)) {
    Drupal.RefreshLess.classes = {};
  }

  /**
   * RefreshLess Turbo adapter class.
   */
  Drupal.RefreshLess.classes.Turbo = class {

    /**
     * Behaviours class instance; responsible for attaching/detaching.
     */
    #Behaviours;

    /**
     * drupalSettings updater class instance.
     */
    #DrupalSettings;

    /**
     * Whether the current response was a redirect.
     *
     * Turbo doesn't provide a clean way to detect this in a
     * 'turbo:before-render' or 'turbo:render' event listener so we need to
     * track whether the most recent response was the result of a redirect to
     * avoid triggering a double set of behaviour detach, settings update, and
     * behaviour attach.
     *
     * @type {Boolean}
     *
     * @see https://www.drupal.org/project/refreshless/issues/3397466
     *   Documents the problem and potential solutions.
     *
     * @see this.#skipRender
     *
     * @todo Rework this once/if Turbo implements a flag or something we can
     *   check during rendering.
     *
     * @todo Can we rework this into a self-contained class or other separate
     *   JavaScript file that listens to events and cancels them itself?
     */
    #isRedirected = false;

    /**
     * Wether to skip the next render, i.e. detach, settings update, and attach.
     *
     * We need a second flag in addition to this.#isRedirected to ensure that we
     * do still detach behaviours even on a redirect navigation, inelegant
     * though it may be.
     *
     * @type {Boolean}
     *
     * @see this.#isRedirected
     */
    #skipRender = false;

    constructor(Behaviours, DrupalSettings) {

      this.#Behaviours = new Behaviours();

      this.#DrupalSettings = new DrupalSettings();

      this.#bindEventHandlers();

      // Remove the reload reason cookie if it exists. This should only occur
      // after a reload when this class is instantiated.
      Cookies.remove(reloadReasonCookieName);

    }

    /**
     * Bind all of our event handlers.
     */
    #bindEventHandlers() {

      /**
       * Reference to the current instance.
       *
       * @type {Turbo}
       */
      const that = this;

      document.documentElement.addEventListener(
        'turbo:before-render',
      async function(
        event,
      ) {
        await that.#beforeRenderHandler(event);
      });

      document.documentElement.addEventListener('turbo:render', function(
        event,
      ) {
        that.#renderHandler(event);
      });

      document.documentElement.addEventListener('turbo:reload', function(
        event,
      ) {

        // Set a cookie containing the reason for the reload so that the
        // back-end can identify this as a Turbo reload. Turbo initiates the
        // reload after triggering this event, so this cookie will be included
        // in the request.
        Cookies.set(
          reloadReasonCookieName,
          event.detail.reason,
          {
            path:     drupalSettings.path.baseUrl,
            sameSite: 'strict',
            secure:   true,
          },
        );

        // Remove the page state cookie before the reload to avoid potentially
        // incorrect libraries on the subsequent page. The back-end should
        // account for this, so this is just in case there's an error or
        // unhandled case in the back-end logic as a second layer to prevent
        // that.
        Cookies.remove(pageStateCookieName);

        console.info('%cRefreshLess%c: Turbo performed a full page load; reason: %c%s%c',
          'font-style: italic', 'font-style: normal',
          'font-family: monospace', event.detail.reason, 'font-family: revert',
        );

      });

      document.documentElement.addEventListener(
        'turbo:before-fetch-request',
      function(event) {
        that.#beforeFetchHandler(event);
      });

    }

    /**
     * 'turbo:before-render' event handler.
     *
     * @param {CustomEvent} event
     *   The event object generated by Turbo.
     *
     * @see this.#beforeRenderTasks()
     *
     * @see https://turbo.hotwired.dev/reference/events
     */
    async #beforeRenderHandler(event) {

      // Don't do anything if this is a cached preview being shown.
      if (typeof $(event.target).attr('data-turbo-preview') !== 'undefined') {
        return;
      }

      // If told to skip this render, set the variable back to false and return.
      if (this.#skipRender === true) {

        this.#skipRender = false;

        return;

      }

      try {

        await this.#beforeRenderTasks(event);

      } catch (error) {

        console.error(error);

        // If anything threw an exception somewhere in our event handler, ensure
        // that Turbo resumes rendering as a failsafe.
        return await event.detail.resume();

      }

    }

    /**
     * 'turbo:before-render' event tasks.
     *
     * This triggers the following RefreshLess events:
     *
     * - 'refreshless:detach': triggered when existing behaviours have been
     *   detached from the page.
     *
     * @param {CustomEvent} event
     *   The event object generated by Turbo.
     *
     * @see https://turbo.hotwired.dev/reference/events
     */
    async #beforeRenderTasks(event) {

      // Pause Turbo rendering.
      event.preventDefault();

      /**
       * Zero or more render delaying promises.
       *
       * This functions as a stack that multiple event subscribers push on to,
       * with rendering only resuming once all Promises are resolved or
       * rejected.
       *
       * @type {Promise[]}
       */
      let renderDelayPromises = [];

      const beforeRenderEvent = new CustomEvent(
        'refreshless:before-render', {detail: {
          turboEvent: event,
          delay: function(executor) {

            const promise = new Promise(executor);

            renderDelayPromises.push(promise);

            return promise;

          },
        },
      });

      document.documentElement.dispatchEvent(beforeRenderEvent);

      /**
       * Reference to the current instance.
       *
       * @type {Turbo}
       */
      const that = this;

      await Promise.allSettled(renderDelayPromises);

      await that.#Behaviours.detach(
        document.documentElement,
      );

      const detachEvent = new CustomEvent('refreshless:detach', {detail: {
        turboEvent: event,
      }});

      document.documentElement.dispatchEvent(detachEvent);

      // Resume Turbo rendering.
      return await event.detail.resume();

    }

    /**
     * 'turbo:render' event handler.
     *
     * @param {CustomEvent} event
     *   The event object generated by Turbo.
     *
     * @see this.#renderTasks()
     *
     * @see https://turbo.hotwired.dev/reference/events
     */
    #renderHandler(event) {

      // Don't do anything if this is a cached preview being shown.
      if (typeof $(event.target).attr('data-turbo-preview') !== 'undefined') {
        return;
      }

      // If this render happens right after a redirect, skip the next render.
      if (this.#isRedirected === true) {

        this.#isRedirected = false;

        this.#skipRender = true;

        return;

      }

      this.#renderTasks(event);

    }

    /**
     * 'turbo:render' event tasks.
     *
     * This triggers the following RefreshLess event:
     *
     * - 'refreshless:drupal-settings-update': triggered when the drupalSettings
     *   global has been updated with new settings from the response.
     *
     * - 'refreshless:attach': triggered when behaviours have been attached to
     *   the new page.
     *
     * @param {CustomEvent} event
     *   The event object generated by Turbo.
     *
     * @see https://turbo.hotwired.dev/reference/events
     */
    #renderTasks(event) {

      /**
       * Reference to the current instance.
       *
       * @type {Turbo}
       */
      const that = this;

      this.#DrupalSettings.update(event.target).then(function(settings) {

        const settingsEvent = new CustomEvent(
          'refreshless:drupal-settings-update', {detail: {
            turboEvent: event,
            settings:   settings,
          }},
        );

        document.documentElement.dispatchEvent(settingsEvent);

        console.debug('%cRefreshLess%c: %cdrupalSettings%c has been updated',
          'font-style: italic', 'font-style: normal',
          'font-weight: bold', 'font-weight: normal',
        );

      }).then(function() {

        return that.#Behaviours.attach(event.target);

      }).then(function() {

        const attachEvent = new CustomEvent('refreshless:attach', {detail: {
          turboEvent: event,
        }});

        document.documentElement.dispatchEvent(attachEvent);

      });

    }

    /**
     * 'turbo:before-fetch-request' event handler.
     *
     * @param {CustomEvent} event
     *   The event object generated by Turbo.
     */
    #beforeFetchHandler(event) {

      // Add a header indicating that this is a RefreshLess Turbo request.
      event.detail.fetchOptions.headers[turboRequestHeaderName] = 'true';

      // Detect if this fetch request is being made to prefetch & return if so.
      //
      // @see https://github.com/hotwired/turbo/pull/1101
      //   Added InstantClick behaviour to Turbo 8.0.0 that prefetches a link on
      //   mouseover and touchstart by default.
      if (
        'X-Sec-Purpose' in event.detail.fetchOptions.headers &&
        event.detail.fetchOptions.headers['X-Sec-Purpose'] === 'prefetch'
      ) {

        console.debug('%cRefreshLess%c: prefetching %c%s',
          'font-style: italic', 'font-style: normal',
          'font-weight: bold', event.detail.url.pathname,
        );

        return;

      }

      /**
       * Whether the current request is a form submit request.
       *
       * Note that we're not checking if this is a POST request because GET
       * requests are also made in some cases, notably Views exposed forms when
       * clicking the Reset button.
       *
       * @type {Boolean}
       */
      let isFormSubmit = false;

      /**
       * 'turbo:submit-start' event handler.
       *
       * This sets isFormSubmit to true if triggered, letting us know that this
       * was a form submit fetch request.
       *
       * @param {CustomEvent} event
       */
      function submitEventHandler(event) {

        isFormSubmit = true;

      };

      document.documentElement.addEventListener(
        'turbo:submit-start', submitEventHandler,
      );

      /**
       * The requested path.
       *
       * @type {String}
       */
      const requestPath = event.detail.url.pathname;

      /**
       * Reference to the current instance.
       *
       * @type {Turbo}
       */
      const that = this;

      // Add a one off event listener for the next response to determine if this
      // was a redirect.
      document.documentElement.addEventListener(
        'turbo:before-fetch-response',
      function(event) {

        // Remove the submit event handler so that we don't keep adding copies
        // of it on each fetch request. This is necessary to do this way rather
        // than by attaching it with {once: true} because if this fetch request
        // wasn't a form submit, the event handler won't be removed because it
        // wouldn't be triggered at all until a later form submit.
        document.documentElement.removeEventListener(
          'turbo:submit-start', submitEventHandler,
        );

        /**
         * The response path.
         *
         * @type {String}
         */
        const responsePath = new URL(
          event.detail.fetchResponse.response.url,
        ).pathname;

        // Turbo does not perform a double render when submitting a form, so log
        // it as such and don't intervene on that redirect as that would cause
        // the destination to not have behaviours attached.
        if (isFormSubmit === true) {

          console.debug('%cRefreshLess%c: form submit to %c%s',
            'font-style: italic', 'font-style: normal',
            'font-weight: bold', responsePath,
          );

          return;

        }

        // If this was not a redirect, log it as a normal navigation and return.
        if (event.detail.fetchResponse.response.redirected === false) {

          console.debug('%cRefreshLess%c: navigated to %c%s',
            'font-style: italic', 'font-style: normal',
            'font-weight: bold', responsePath,
          );

          return;

        }

        console.debug('%cRefreshLess%c: redirected %c%s%c → %c%s',
          'font-style: italic', 'font-style: normal',
          'font-weight: bold', requestPath, 'font-weight: normal',
          'font-weight: bold', responsePath,
        );

        that.#isRedirected = true;

      }, {once: true});

    }

  }

  if (!('Turbo' in Drupal.RefreshLess)) {

    Drupal.RefreshLess.Turbo = new Drupal.RefreshLess.classes.Turbo(
      Drupal.RefreshLess.classes.TurboBehaviours,
      Drupal.RefreshLess.classes.TurboDrupalSettings,
    );

  }

  // This adds a session cookie containing the libraries, theme name, and theme
  // token from drupalSettings.ajaxPageState so that they're sent with every
  // Turbo request.
  Drupal.behaviors.refreshlessTurboPageState = {

    attach(context, settings) {

      once('refreshless-turbo-page-state', 'html').forEach(function() {

        const ajaxPageState = drupalSettings.ajaxPageState;

        const refreshlessPageState = {
          libraries:    ajaxPageState.libraries,
          theme:        ajaxPageState.theme,
          theme_token:  ajaxPageState.theme_token,
        };

        Cookies.set(
          pageStateCookieName,
          JSON.stringify(refreshlessPageState),
          {
            path:     drupalSettings.path.baseUrl,
            sameSite: 'strict',
            secure:   true,
          },
        );

      });

    },

    detach(context, settings, trigger) {

      if (trigger !== 'unload') {
        return;
      }

      // We don't strictly need to remove the cookie, but it has no use after
      // the request has been sent so we might as well clean up after ourselves.
      // We do need to once.remove() however, to ensure we attach again on the
      // next page and thus update the cookie contents.
      once.remove('refreshless-turbo-page-state', 'html').forEach(function() {
        Cookies.remove(pageStateCookieName);
      });

    },

  };

})(Drupal, drupalSettings, jQuery, Cookies);
