/**
 * @file
 * Sort stylesheet <link> elements to fix Turbo breaking CSS specificity.
 *
 * @see https://www.drupal.org/project/refreshless/issues/3399314
 *   CSS specificity issue.
 *
 * @todo Contribute an upstream fix to Turbo and remove this if no longer
 *   needed. This still results in some brief layout shifts since it doesn't
 *   seem like Turbo exposes a way to act on the <head> contents right as
 *   they're added, which happens even before the 'turbo:before-render' event is
 *   triggered. A MutationObserver could be used to start sorting the elements
 *   as soon as they start getting added, but there's no guarantee its callback
 *   would get all of them in one invocation so we would need to provide the
 *   order via some other means - e.g. drupalSettings.
 */
(function (Drupal, $, once) {

  'use strict';

  /**
   * The once() identifier for our behaviour.
   *
   * @type {String}
   */
  const onceName = 'refreshless-turbo-stylesheet-sorter';

  /**
   * The order attribute name added to <link> elements by the back-end.
   *
   * @type {String}
   *
   * @todo Pass this from the back-end via drupalSettings instead of hard-coding
   *   it here.
   */
  const weightAttributeName = 'data-refreshless-turbo-order';

  function sortEventHandler(event) {

    const $elements = $(
      `head link[rel="stylesheet"][${weightAttributeName}]`, this,
    );

    /**
     * True if the sort order was changed, false otherwise.
     *
     * This prevents unnececessary DOM manipulation if no sorting occurred.
     *
     * @type {Boolean}
     */
    let wasSorted = false;

    /**
     * <link> elements sorted by the order property generated by the back-end.
     *
     * @type {HTMLLinkElement[]}
     *
     * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/sort
     */
    const sorted = $elements.toArray().sort(function(a, b) {

      const aWeight = Number.parseInt($(a).attr(weightAttributeName));

      const bWeight = Number.parseInt($(b).attr(weightAttributeName));

      if (aWeight < bWeight) {

        wasSorted = true;

        return -1;

      } else if (aWeight > bWeight) {

        wasSorted = true;

        return 1;

      }

      return 0;

    });

    if (wasSorted === false) {
      return;
    }

    // Append all the sorted <link> elements. This will be done in the newly
    // sorted order.
    $('head', this).append(sorted);

  };

  // Note that we're not using a behaviour here since it doesn't make sense to
  // detach and re-attach on every page. This doesn't seem to make a difference
  // in terms of layout shifts, but it reduces work the browser has to do, even
  // if minor.
  $(once(onceName, 'html')).on('turbo:before-render', sortEventHandler);

})(Drupal, jQuery, once);
