<?php

declare(strict_types=1);

namespace Drupal\refreshless_turbo\Hooks;

use Drupal\Core\Access\CsrfTokenGenerator;
use Drupal\Core\Asset\AttachedAssetsInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\hux\Attribute\Alter;
use Drupal\hux\Attribute\Hook;
use Drupal\refreshless_turbo\Service\RefreshlessTurboKillSwitchInterface;
use function in_array;

/**
 * JavaScript hook implementations.
 */
class Javascript {

  /**
   * Hook constructor; saves dependencies.
   *
   * @param \Drupal\Core\Access\CsrfTokenGenerator $csrfTokenGenerator
   *   The CSRF token generator.
   *
   * @param \Drupal\refreshless_turbo\Service\RefreshlessTurboKillSwitchInterface $killSwitch
   *   The RefreshLess Turbo kill switch service.
   *
   * @param \Drupal\Core\Extension\ThemeHandlerInterface $themeHandler
   *   The theme handler.
   *
   * @param \Drupal\Core\Theme\ThemeManagerInterface $themeManager
   *   The theme manager.
   */
  public function __construct(
    protected readonly CsrfTokenGenerator     $csrfTokenGenerator,
    protected readonly RefreshlessTurboKillSwitchInterface $killSwitch,
    protected readonly ThemeHandlerInterface  $themeHandler,
    protected readonly ThemeManagerInterface  $themeManager,
  ) {}

  #[Alter('js')]
  /**
   * Move all JavaScript to the <head> as recommended for Hotwire Turbo.
   *
   * JavaScript found in the <body> will be executed each time that page is
   * loaded whereas JavaScript linked in the <head> will only be executed on a
   * full page load. Since all Drupal JavaScript should be using behaviours,
   * they should all correctly re-attach on every Turbo visit.
   *
   * @see \hook_js_alter()
   */
  public function alter(
    array &$javascript,
    AttachedAssetsInterface $assets, LanguageInterface $language,
  ): void {

    if ($this->killSwitch->triggered()) {
      return;
    }

    foreach ($javascript as $filePath => &$settings) {

      // Skip items that are already set to the header as they may need to be
      // render blocking rather than deferred.
      if ($settings['scope'] === 'header') {
        continue;
      }

      $settings['scope'] = 'header';

      if (!isset($settings['attributes']['defer'])) {
        $settings['attributes']['defer'] = true;
      }

    }

  }

  #[Alter('js')]
  /**
   * Enable aggregation for our JavaScript if core supports naming groups.
   *
   * If core does not support named aggregation groups, the 'group' value will
   * never be our string but a numeric value that core sets; if it's the numeric
   * value, we don't do anything and allow the preprocess: false in our library
   * definition to be used as a fallback.
   *
   * Note that \hook_library_info_alter() seems to be too early to detect this
   * as it'll always have our string group; core seems to override the group
   * value sometime between invoking that hook and \hook_js_alter().
   *
   * Also note that this should always be done regardless of whether or not this
   * is a RefreshLess request, nor if the kill switch has been triggered.
   *
   * @see \hook_js_alter()
   */
  public function refreshLessTurboGroupPreprocess(
    array &$javascript,
    AttachedAssetsInterface $assets, LanguageInterface $language,
  ): void {

    foreach ($javascript as $path => &$values) {

      if (!in_array($values['group'], [
        'refreshless-turbo', 'refreshless-turbo-reload',
      ])) {
        continue;
      }

      // Named aggregation groups are supported, so enable aggregation for this
      // group.
      $values['preprocess'] = true;

    }

  }

  #[Hook('js_settings_build')]
  /**
   * Output 'ajaxPageState' to enable additive JavaScript aggregation.
   *
   * @see https://www.drupal.org/project/refreshless/issues/3414538
   *   Details regarding additive JavaScript aggregation and why it's needed.
   */
  public function outputPageState(
    array &$settings, AttachedAssetsInterface $assets,
  ): void {

    // Don't do anything if ajaxPageState is already present or the kill switch
    // has been triggered.
    if (
      isset($settings['ajaxPageState']) ||
      $this->killSwitch->triggered()
    ) {
      return;
    }

    /** @var string The active theme's machine name. */
    $activeThemeKey = $this->themeManager->getActiveTheme()->getName();

    $settings['ajaxPageState'] = [
      // \system_js_settings_build() builds 'libraries' for us if
      // 'ajaxPageState' is present.
      'theme' => $activeThemeKey,
    ];

    // Only output the theme token if the active theme is different from the
    // default theme.
    //
    // @see \system_js_settings_alter()
    if ($activeThemeKey !== $this->themeHandler->getDefault()) {

      $settings['ajaxPageState'][
        'theme_token'
      ] = $this->csrfTokenGenerator->get($activeThemeKey);

    }

  }

}
