<?php

declare(strict_types=1);

namespace Drupal\Tests\refreshless_turbo\FunctionalJavascript;

use Behat\Mink\Exception\ElementHtmlException;
use Behat\Mink\Exception\ExpectationException;
use Drupal\FunctionalJavascriptTests\WebDriverWebAssert;

/**
 * WebDriverWebAssert class extended with Turbo assertion methods.
 *
 * @see \Drupal\Tests\refreshless_turbo\FunctionalJavascript\TurboAssertTrait::assertSession()
 */
class TurboWebDriverAssert extends WebDriverWebAssert {

  /**
   * Name of the attribute set on the <html> element to track Turbo persisting.
   */
  protected const TURBO_PERSIST_ATTR = 'data-refreshless-turbo-test';

  /**
   * Assert that Turbo is present on the page.
   */
  public function assertTurboIsPresent(): void {

    $result = $this->session->getDriver()->wait(10000, <<<JS
      'Turbo' in window;
    JS);

    $this->assert(
      $result === true, 'Turbo is not present in the window object.',
    );

  }

  /**
   * Waits for a Turbo request to complete.
   *
   * @param int $timeout
   *   (Optional) Timeout in milliseconds, defaults to 10000.
   *
   * @param string $message
   *   (optional) A message for exception.
   *
   * @throws \RuntimeException
   *   When the request is not completed. If left blank, a default message will
   *   be displayed.
   *
   * @see \Drupal\FunctionalJavascriptTests\JSWebAssert::assertWaitOnAjaxRequest()
   *   Loosely based on this core assert method.
   */
  public function assertWaitOnTurboRequest(
    $timeout = 10000, $message = 'Unable to complete Turbo request.',
  ) {

    // Wait for a very short time to allow page state to update after clicking.
    \usleep(5000);

    $condition = <<<JS
      (function() {

        return typeof jQuery(
          document.documentElement,
        ).attr('aria-busy') === 'undefined';

      }());
    JS;

    $result = $this->session->wait($timeout, $condition);

    if ($result !== true) {
      throw new \RuntimeException($message);
    }

  }

  /**
   * Start tracking whether the document persists after Turbo navigation.
   *
   * @see $this->assertTurboPersisted()
   *
   * @see $this->assertTurboNotPersisted()
   */
  public function startTurboPersist(): void {

    $attributeName = self::TURBO_PERSIST_ATTR;

    $this->session->evaluateScript(<<<JS
      jQuery('html').attr('{$attributeName}', true);
    JS);

  }

  /**
   * Assert that the document has persisted after a Turbo navigation.
   *
   * @see $this->startTurboPersist()
   *   Must be called to start tracking persist state before asserting.
   *
   * @see $this->assertTurboNotPersisted()
   *   Inverse of this method.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   *   If it doesn't appear the page has persisted, i.e. a full page load has
   *   occurred.
   */
  public function assertTurboPersisted(): void {

    try {

      $this->elementAttributeExists('css', 'html', self::TURBO_PERSIST_ATTR);

    } catch (ElementHtmlException $exception) {

      throw new ExpectationException(
        'Either a full, non-Turbo page load has occurred, or TurboWebDriverAssert::startTurboPersist() was not called beforehand.',
        $this->session->getDriver(),
      );

    }

  }

  /**
   * Assert that the document has not persisted due to a Turbo navigation.
   *
   * @see $this->startTurboPersist()
   *   Must be called to start tracking persist state before asserting.
   *
   * @see $this->assertTurboPersisted()
   *   Inverse of this method.
   *
   * @throws \Behat\Mink\Exception\ExpectationException
   *   If it appears the page has persisted, i.e. a full page load has not
   *   occurred.
   */
  public function assertTurboNotPersisted(): void {

    try {

      $this->elementAttributeNotExists('css', 'html', self::TURBO_PERSIST_ATTR);

    } catch (ElementHtmlException $exception) {

      throw new ExpectationException(
        'Either a full page load has not occurred, or TurboWebDriverAssert::startTurboPersist() was called since the last full page load.',
        $this->session->getDriver(),
      );

    }

  }

}
